#define USECHRONO
#include <cassert>
#include <climits>
#include <cstdlib>
#include <stack>
#include <Eigen/Dense>
#include "eval.hpp"

using namespace aed;
using namespace std;
using namespace Eigen;

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
typedef int (*mapfun_t)(int);
void separate(set<int> &S,mapfun_t f,map<int,set<int> > &f2x) {
  f2x.clear();
  for (auto x:S) {
    int y = f(x);
    f2x[y].insert(x);
  }
}

class graph_t {
public:
  virtual int size()=0;
  virtual int dist(int i,int j)=0;
};

bool next_path(vector<int> &path) {
  int N = path.size();
  for (int j=N-1; j>=0; j--) {
    if (path[0]==j) {
      for (int k=0; k<j; k++) path[k] = path[k+1];
      path[j] = j;
    } else {
      int k;
      for (k=0; k<N-1; k++) 
	if (path[k]==j) break;
      path[k] = path[k-1];
      path[k-1] = j;
      return 1;
    }
  }
  return 0;
}

int min_path(graph_t &G,vector<int> &pathmin) {
  int N = G.size();
  vector<int> path(N);
  for (int j=0; j<N; j++) path[j] = j;
  int lenmin=-1;
  while (1) {
    int len=0;
    for (int j=0; j<N; j++) {
      int k = (j+1)%N;
      len += G.dist(path[j],path[k]);
    }
    if (lenmin<0 || len<lenmin) {
      lenmin=len;
      pathmin = path;
    }
    if (!next_path(path)) break;
  }
  return lenmin;
}

class circle_t : public graph_t {
public:
  int N;
  int size() { return N; }
  int dist(int i,int j) {
    if (i==(j+1)%N || j==(i+1)%N) return 1;
    else return 10000;
  }
};

class full_t : public graph_t {
public:
  int N;
  int size() { return N; }
  int dist(int i,int j) { return 1; }
};

class grid_t : public graph_t {
public:
  int Nx,Ny;
  int size() { return Nx*Ny; }
  void gjk(int n,int &j,int &k) {
    j = n/Nx;
    k = n%Nx;
  }
  int dist(int n1,int n2) {
    int j1,j2,k1,k2;
    gjk(n1,j1,k1);
    gjk(n2,j2,k2);
    return abs(j2-j1)+abs(k2-k1);
  }
};

void ppath(const vector<int> &path) {
  for (auto q:path) cout << q << " ";
  cout << endl;
}

void check_min_path() {
  cout << "-------------------" << endl
       << "CHECK MINPATH" << endl;
  vector<int> path;
  int minlen;

  circle_t G;
  G.N=4;
  minlen = min_path(G,path);
  cout << "circle graph(4): ";
  ppath(path);
  cout << "min length: " << minlen << endl;

  G.N=7;
  minlen = min_path(G,path);
  cout << "circle graph(7): ";
  ppath(path);
  cout << "min length: " << minlen << endl;
  
  full_t G2;
  G2.N=4;
  minlen = min_path(G2,path);
  cout << "full graph: ";
  ppath(path);
  cout << "min length: " << minlen << endl;

  grid_t G3;
  G3.Nx=3;
  G3.Ny=3;
  minlen = min_path(G3,path);
  cout << "grid graph 3x3: ";
  ppath(path);
  cout << "min length: " << minlen << endl;

  G3.Nx=4;
  G3.Ny=2;
  minlen = min_path(G3,path);
  cout << "grid graph 4x2: ";
  ppath(path);
  cout << "min length: " << minlen << endl;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void mkortho(MatrixXd &Q) {
  int n = Q.rows();
  MatrixXd Id = MatrixXd::Identity(n,n),DQ,Z;
  int j=0;
  while (1) {
    DQ = Id-Q*Q.transpose();
    double erro = DQ.norm();
    // cout << j << ": " << erro << endl;
    if (erro<1e-6) break;
    Z = Q.transpose();
    Z = Z.inverse();
    DQ = 0.5*DQ*Z;
    Q += DQ; j++;
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void check_mkortho() {
  cout << "-------------------" << endl
       << "CHECK MKORTHO" << endl;
  for (int z=0; z<10; z++) {
    int N=3+rand()%4;
    MatrixXd Q(N,N),Q0;
    for (int j=0; j<N; j++) {
      for (int k=0; k<N; k++) 
        Q(j,k) = 0.1*drand();
      Q(j,j) += 1.0;
    }
    Q0=Q;
    // cout << Q << endl;
    mkortho(Q);
    // cout << Q << endl;
    MatrixXd DQ = Q-Q0;
    double adq = DQ.norm();
    MatrixXd Id = MatrixXd::Identity(N,N);
    MatrixXd E = Q*Q.transpose()-Id;
    double erro = E.norm();
    cout << "correction: " << adq << ", erro " << erro << endl;
  }
}  

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main() {
  // Para chequear ejercicio 1
  int vrbs=0;
  Eval ev;
  ev.eval<1>(separate,vrbs);

  // Para chequear ejercicio 2
  // Para probar el recorrido de los caminos
  // print_all_paths();
  check_min_path();

  // Para chequear ejercicio 3
  check_mkortho();

  return 0;
}
