#define USECHRONO
#include <cassert>
#include <climits>
#include <cstdlib>
#include <Eigen/Dense>
#include "eval.hpp"

using namespace aed;
using namespace std;

// Calcular la superficie de una esfera
using namespace Eigen;
using namespace std;

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class surface_t {
public:
  virtual void generate(VectorXd &uv,VectorXd &xyz)=0;
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
// Computes the are of a triangle in 3D
double areatri(VectorXd &y0,VectorXd &y1,VectorXd &y2) {
  Vector3d a=y1-y0, b=y2-y0,nor;
  nor = a.cross(b);
  return nor.norm()/2.0;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
double area(surface_t &surf,int N=100,
            double umin=0.0,double umax=1.0,
            double vmin=0.0,double vmax=1.0) {
  // N is the number of segments in each axis
  // Positions of the vertex of the quadrangle.
  // XC will be the center.
  const int ndim=3;
  VectorXd x1(ndim),x2(ndim),x3(ndim),x4(ndim),xc(ndim);
  // Accumulator for the area
  // C is the radius in Z direction
  double Area=0.0;
  // Compute the areas for the projections of 1/4th of the
  // plane at z=1, x=1, and y=1.
  VectorXd uv(2);
  double hu=(umax-umin)/N;
  double hv=(vmax-vmin)/N;
  // Loop over the quadrangles
  for (int j=0; j<N; j++) {
    double u0=umin+j*hu;
    for (int k=0; k<N; k++) {
      double v0=vmin+k*hv;
      uv << u0,v0; surf.generate(uv,x1);
      uv << u0+hu,v0; surf.generate(uv,x2);
      uv << u0+hu,v0+hv; surf.generate(uv,x3);
      uv << u0,v0+hv; surf.generate(uv,x4);
      xc = (x1+x2+x3+x4)/4.0;
      // Compute the area of 4 triangles composed of each
      // side of the quadrangle and the center of the sphere
      Area += areatri(x1,x2,xc);
#define ESHV(x) cout << #x << endl << x << endl
      // ESHV(x1);
      // ESHV(x2);
      // ESHV(x3);
      // ESHV(x4);
      // ESHV(xc);
      Area += areatri(x2,x3,xc);
      Area += areatri(x3,x4,xc);
      Area += areatri(x4,x1,xc);
    }
  }
  return Area;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class plane_t : public surface_t {
public:
  double c,a,b;
  void generate(VectorXd &uv,VectorXd &xyz) {
    xyz(0) = uv(0);
    xyz(1) = uv(1);
    xyz(2) = c+a*uv(0)+b*uv(1);
  };
  plane_t(double a_=0.0,double b_=0.0,double c_=0.0)
    : c(c_), a(a_),b(b_) {}
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class ellipsoid_t : public surface_t {
public:
  double a,b,c;
  void generate(VectorXd &uv,VectorXd &xyz) {
    double theta=uv(0),phi=uv(1);
    xyz(0) = a*cos(theta)*cos(phi);
    xyz(1) = b*cos(theta)*sin(phi);
    xyz(2) = c*sin(theta);
  };
  ellipsoid_t() : a(1.0), b(1.0), c(1.0) { }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class cone_t : public surface_t {
public:
  double R,theta;
  void generate(VectorXd &uv,VectorXd &xyz) {
    double rho=R*uv(0),phi=uv(1);
    xyz(0) = rho*cos(phi);
    xyz(1) = rho*sin(phi);
    xyz(2) = rho/tan(theta);
  };
  cone_t() : R(NAN),theta(NAN) { }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class catenoid_t : public surface_t {
public:
  double a,c;
  void generate(VectorXd &uv,VectorXd &xyz) {
    double z=uv(0),phi=uv(1);
    double rho = a*cosh(z/c);
    xyz << rho*cos(phi),rho*sin(phi),z;
  };
  catenoid_t() : a(NAN), c(NAN) { }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class objfun_t {
public: 
  virtual double eval(VectorXd &X)=0;
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class optimizer_t {
public:
  double omega,tol;
  int niter,vrbs;
  optimizer_t() : omega(0.1),tol(1e-3),niter(100), vrbs(1) {}
  double optimize(objfun_t &F,VectorXd &X) {
    double f;
    for (int it=0; it<niter; it++) {
      if (vrbs) {
        cout << "iter,X " << it;
        ESHV(X);
      }
      f = F.eval(X);
      int n = X.size();
      VectorXd gradf(n);
      double epsln = 1e-3;
      for (int j=0; j<n; j++) {
        VectorXd X1;
        X1 = X; X1(j) += epsln; double fp = F.eval(X1);
        X1 = X; X1(j) -= epsln; double fm = F.eval(X1);
        gradf(j) = (fp-fm)/(2*epsln);
      }
      // ESHV(gradf);
      X -= omega*gradf;
      if (gradf.norm()<tol) break;
      if (vrbs) cout << "iter,F,||g|| " << it << " "
           << f << " " << gradf.norm() << endl;
    }
    return f;
  }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class quadfun_t : public objfun_t {
 public:
  VectorXd Xopt;
  double c;
  double eval(VectorXd &X) {
    VectorXd dx = X-Xopt;
    return c*dx.squaredNorm();
  }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class genrevol_t : public surface_t {
public:
  double R,L;
  VectorXd X;
  void generate(VectorXd &uv,VectorXd &xyz) {
    double rho=R;
    int n=X.size();
    double phi=uv(0),z=uv(1),zeta=M_PI*z/(2.0*L);
    for (int j=0; j<n; j++)
      rho -= X(j)*cos((2*j+1)*zeta);
    // printf("zeta %f, rho %f\n",zeta,rho);
    xyz << rho*cos(phi),rho*sin(phi),z;
    // printf("xyz %f %f %f\n",xyz(0),xyz(1),xyz(2));
  };
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class area_surf_t : public objfun_t {
 public:
  genrevol_t surf;
  double eval(VectorXd &X) {
    surf.X = X;
    double A = 8*area(surf,10,0,-M_PI/2,0,surf.L);
    return A;
  }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class snell_t : public objfun_t {
public:
  vector<double> d,c;
  double L;
  double eval(VectorXd &X) {
    int n=d.size();
    assert(int(c.size())==n);
    assert(X.size()==n-1);
    double F=0.0;
    VectorXd XX(n);
    XX.segment(0,n-1) = X;
    XX(n-1) = L-X.sum();
    for (int j=0; j<n; j++) 
      F += sqrt(d[j]*d[j]+XX(j)*XX(j))/c[j];
    return F;
  }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main() {
  const double pi=M_PI;
#if 0
  double a=1.0,b=1.0,c=10.0;
  plane_t plane(a,b,c);
  double Area = sqrt(1.0+a*a+b*b);
  cout << "a,b,c, Area plane (num/exact): "
       << plane.a << " " << plane.b << " " << plane.c << " "
       << area(plane) << " " << Area << endl;
#elif 0
  ellipsoid_t ellips;
  cout << "Sphere surface (num/ex): "
       << 8*area(ellips,100,0,pi/2,0,pi/2)
       << " " << 4*M_PI << endl;
  ellips.c = 0.5;
  cout << "ellipsoid (" << ellips.a << "," << ellips.b << "," << ellips.c
       << ") area: " << 8*area(ellips,100,0,pi/2,0,pi/2) << endl;
  ellips.c = 2.0;
  cout << "ellipsoid (" << ellips.a << "," << ellips.b << "," << ellips.c
       << ") area: " << 8*area(ellips,100,0,pi/2,0,pi/2) << endl;
#elif 0
  cone_t cone;
  cone.R = 1.0;
  cone.theta = pi/4.0;
  // double RR=sqrt(cone.R*cone.R+
  cout << "Cone surface (num/ex): "
       << area(cone,0,1,-pi,pi)
       << " " << 4*M_PI << endl;
  ellips.c = 0.5;
  cout << "ellipsoid (" << ellips.a << "," << ellips.b << "," << ellips.c
       << ") area: " << 8*area(ellips,100,0,pi/2,0,pi/2) << endl;
  ellips.c = 2.0;
  cout << "ellipsoid (" << ellips.a << "," << ellips.b << "," << ellips.c
       << ") area: " << 8*area(ellips,100,0,pi/2,0,pi/2) << endl;
#elif 0
  catenoid_t cat;
  double L=0.6,R=1;
  double cmin=0.1,cmax=2;
  int N=20,Nseg=400;
  double acyl=4.0*pi*R*L;
  for (int j=0; j<=N; j++) {
    cat.c = cmin+double(j)/N*(cmax-cmin);
    cat.a = R/cosh(L/cat.c);
    printf("c %f, a(neck rat) %f, area %f\n",
           cat.c,cat.a,8.0*area(cat,Nseg,0,L,0,pi/2)/acyl);
  }
#elif 0
  optimizer_t opt;
  opt.niter = 100;
  opt.omega = 0.1;
  quadfun_t qf;
  int n=1;
  qf.Xopt.resize(n);
  qf.Xopt.fill(1.0);
  qf.c=2.0;
  VectorXd X=0*qf.Xopt;
  opt.optimize(qf,X);
#elif 1
  area_surf_t F;
  F.surf.R=1.0;
  F.surf.L=0.6;
  int N=10;
  double amax = 0.4;
  VectorXd X(1);
  for (int j=0; j<=N; j++) {
    X(0) = -double(j)/N*amax;
    printf("j %d, a %f, F %f\n",j,X(0),F.eval(X));
  }
  // Number of parameters
  optimizer_t opt;
  opt.niter = 1000;
  opt.omega = 0.01;
  opt.vrbs = 0;
  for (int m=1; m<=2; m++) {
    X.resize(m);
    X.fill(0.0);
    opt.optimize(F,X);
    ESHV(X);
    cout << "m " << m << ", Surface " << F.eval(X) << endl;
  }
#elif 0
  snell_t snell;
  snell.d.push_back(1.0);
  snell.d.push_back(1.0);
  snell.c.push_back(1.0);
  snell.c.push_back(2.0);
  snell.L = 1.0;
  VectorXd X(1);
  X(0) = 0.5;
  optimizer_t opt;
  opt.niter = 100;
  opt.omega = 1;
  opt.tol = 1e-7;
  opt.optimize(snell,X);
#else
  // Option not chosen
  assert(0);
#endif
  return 0;
}
