#include "eval.hpp"
#include <cmath>
#include <boost/numeric/odeint.hpp>

using namespace aed;
using namespace std;
using namespace boost::numeric::odeint;

void nproxies(graph_t &G,vector<int> &np,int d) {
  int N = G.size();
  vector<int> distance,path;
  np.resize(N,0);
  for (int j=0; j<N; j++) {
    G.breadth_first(j,path,distance);
    int count=0;
    for (int k=0; k<N; k++)
      count += (distance[k]==d);
    np[j] = count;
  }
  // printf("--\n");
  // for (int j=0; j<N; j++) printf("j %d, np %d\n",j,np[j]);
  sort(np.begin(),np.end());
}

bool isomorph(graph_t &G1,graph_t &G2) {
  if (G1.size() != G2.size()) return false;
  vector<int> np1,np2;
  for (int d=1; d<=2; d++) {
    nproxies(G1,np1,d);
    nproxies(G2,np2,d);
    int N = G1.size();
    // printf("---- d=%d\n",d);
    // for (int j=0; j<N; j++)
    //   printf("j, np1, np2, %d %d %d\n",j,np1[j],np2[j]);
    if (np1!=np2) return false;
  }
  return true;
}

class vector_field_t {
public:
  virtual void field(MatrixXd &x,MatrixXd &V)=0;
};

class shape_t {
public:
  virtual int size()=0;
  virtual void xnode(int j,MatrixXd &x)=0;
};

class uniform_t : public vector_field_t {
public:
  MatrixXd w;
  uniform_t() { w.resize(2,1); }
  void field(MatrixXd &x,MatrixXd &V) { V = w;  }
};

class segment_t : public shape_t {
public:
  int N;
  MatrixXd start,end;
  segment_t() { start.resize(2,1); end.resize(2,1); }
  int size() { return N; }
  void xnode(int j,MatrixXd &x) {
    x = start+(end-start)*double(j)/N;
  }
};

class circle_t : public shape_t {
public:
  int N;
  double R;
  MatrixXd center;
  circle_t() {
    N = 100;
    R = 1.0;
    center.resize(2,1);
    center << 0,0;
  }
  int size() { return N; }
  void xnode(int j,MatrixXd &x) {
    double phi = 2*M_PI*double(j)/N;
    x.resize(2,1);
    x << 0.5+cos(phi), sin(phi);
  }
};
  
typedef double (*flux_t)(shape_t &shape,vector_field_t &vf);

double flux(shape_t &shape,vector_field_t &vf) {
  int nelem=shape.size(), nnod = nelem+1;
  MatrixXd x1(2,1),x2(2,1), dx(2,1),xc(2,1),V(2,1),ds(2,1);
  double F = 0.0;
  for (int j=0; j<nelem; j++)  {
    shape.xnode(j,x1);
    shape.xnode(j+1,x2);
    dx = (x2-x1);
    xc = 0.5*(x1+x2);
    vf.field(xc,V);
    ds(1,0) = -dx(0,0);
    ds(0,0) = dx(1,0);
    F += ds(0,0)*V(0,0)+ds(1,0)*V(1,0);
  }
  return F;
}

class radial_t : public vector_field_t {
public:
  MatrixXd center;
  double intensity;
  radial_t() {
    center.resize(2,1);
    center << 0,0;
    intensity = 1;
  }
  void field(MatrixXd &x,MatrixXd &V) {
    V = intensity*(x-center);
  }
};

class solenoidal_t : public vector_field_t {
public:
  double intensity;
  solenoidal_t() { intensity=1.0; }
  void field(MatrixXd &x,MatrixXd &V) {
    V.resize(2,1);
    V(1,0) = intensity*x(0,0);
    V(0,0) = -intensity*x(1,0);
  }
};

void check_flux(flux_t flux_fun) {
  segment_t segment;
  uniform_t uniform;

  segment.N = 100;
  segment.start << 0,0;
  segment.end << 1,0;
  uniform.w << 0,1;
  printf("flux (segment[1,0], uniform(0,1): %f\n",flux_fun(segment,uniform));

  uniform.w << 1,0;
  printf("flux (segment[1,0], uniform(1,0): %f\n",flux_fun(segment,uniform));

  segment.start << 0,0;
  segment.end << 0,1;
  uniform.w << 1,0;
  printf("flux (segment[0,1], uniform(1,0): %f\n",flux_fun(segment,uniform));

  uniform.w << 0,1;
  printf("flux (segment[0,1], uniform(0,1): %f\n",flux_fun(segment,uniform));

  radial_t radial;
  circle_t circle;
  printf("flux (circle(c=(0,0),R=1), radial(c=(0,0),int=1): %f\n",
         flux_fun(circle,radial));
  circle.N = 1000;
  printf("flux circle(c=(0,0),R=1), radial(c=(0,0),intens=1), N=1000: %f\n",
         flux_fun(circle,radial));

  radial.intensity = 10.0;
  printf("flux circle(c=(0,0),R=1), radial(c=(0,0),intens=100): %f\n",
         flux_fun(circle,radial));

  solenoidal_t sole;
  printf("flux circle(c=(0,0),R=1), sole(c=(0,0),intens=100): %f\n",
         flux_fun(circle,sole));
}

typedef vector< double > vec_t;
class tanks_t {
public:
  double A1,C2,rho,g,C3,H,Q10,T,A2,C4;
  void operator() (const vec_t &x, vec_t &dxdt,double t) {
    double
      h1 = x[0],
      h2= x[1];
    double tau = t-floor(t/T)*T;
    double
      Q1 = Q10*(tau<T/2.0),
      Q2 = C2*rho*h1*g,
      Q3 = C3*(h1-H)*(h1>H),
      Q4 = C4*rho*h2*g;
    dxdt[0] = (Q1-Q2-Q3)/A1;
    dxdt[1] = (Q3-Q4)/A2;
  }
} tanks;

struct observer_t {
  void operator()( const vec_t &x , double t ) {
    printf("%f ",t);
    int N = x.size();
    for (int j=0; j<N; j++) printf("%f ",x[j]);
    printf("\n");
  }
} observer;

void check_ode() {
  vec_t x(2);
  tanks_t tanks;
  tanks.A1 = 1;
  tanks.C2 = 0.5e-5;
  tanks.rho = 1000;
  tanks.g = 10;
  tanks.C3 = 1;
  tanks.H = 1;
  tanks.Q10 = 0.3;
  tanks.T = 40;
  tanks.A2 = 100;
  tanks.C4 = 1e-4;

  runge_kutta4<vec_t> stepper;
  size_t steps = integrate_const(stepper,tanks,x,0.0,500.0,0.1,observer);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main() {
  Eval ev;
  int vrbs=0, seed=123;

  ev.eval1(isomorph,vrbs);
  // ev.evalr1(isomorph,vrbs);
  // check_flux(flux);
  // check_ode();
  
  return 0;
}
